"""Register names from the Huawei inverter."""

MODEL_NAME = "model_name"
SERIAL_NUMBER = "serial_number"
PN = "pn"
FIRMWARE_VERSION = "firmware_version"
SOFTWARE_VERSION = "software_version"
PROTOCOL_VERSION_MODBUS = "protocol_version_modbus"
MODEL_ID = "model_id"
NB_PV_STRINGS = "nb_pv_strings"
NB_MPP_TRACKS = "nb_mpp_tracks"
RATED_POWER = "rated_power"
P_MAX = "P_max"
S_MAX = "S_max"
Q_MAX_OUT = "Q_max_out"
Q_MAX_IN = "Q_max_in"
P_MAX_REAL = "P_max_real"
S_MAX_REAL = "S_max_real"
PRODUCT_SALES_AREA = "product_sales_area"
PRODUCT_SOFTWARE_NUMBER = "product_software_number"
PRODUCT_SOFTWARE_VERSION_NUMBER = "product_software_version_number"
GRID_STANDARD_CODE_PROTOCOL_VERSION = "grid_standard_code_protocol_version"
UNIQUE_ID_OF_THE_SOFTWARE = "unique_id_of_the_software"
NUMBER_OF_PACKAGES_TO_BE_UPGRADED = "number_of_packages_to_be_upgraded"
HARDWARE_FUNCTIONAL_UNIT_CONF_ID = "hardware_functional_unit_conf_id"
SUBDEVICE_SUPPORT_FLAG = "subdevice_support_flag"
SUBDEVICE_IN_POSITION_FLAG = "subdevice_in_position_flag"
FEATURE_MASK_1 = "feature_mask_1"
FEATURE_MASK_2 = "feature_mask_2"
FEATURE_MASK_3 = "feature_mask_3"
FEATURE_MASK_4 = "feature_mask_4"
REALTIME_MAX_ACTIVE_CAPABILITY = "realtime_max_active_capability"
REALTIME_MAX_INDUCTIVE_REACTIVE_CAPACITY = "realtime_max_inductive_reactive_capacity"
OFFERING_NAME_OF_SOUTHBOUND_DEVICE_1 = "offering_name_of_southbound_device_1"
OFFERING_NAME_OF_SOUTHBOUND_DEVICE_2 = "offering_name_of_southbound_device_2"
OFFERING_NAME_OF_SOUTHBOUND_DEVICE_3 = "offering_name_of_southbound_device_3"
HARDWARE_VERSION = "hardware_version"
MONITORING_BOARD_SN = "monitoring_board_sn"
MONITORING_SOFTWARE_VERSION = "monitoring_software_version"
MASTER_DSP_VERSION = "master_dsp_version"
SLAVE_DSP_VERSION = "slave_dsp_version"
CPLD_VERSION = "cpld_version"
AFCI_VERSION = "afci_version"
BUILTIN_PID_VERSION = "builtin_pid_version"
DC_MBUS_VERSION = "dc_mbus_version"
EL_MODULE_VERSION = "el_module_version"
AFCI_2_VERSION = "afci_2_version"
REGKEY = "regkey"
STATE_1 = "state_1"
STATE_2 = "state_2"
STATE_3 = "state_3"
ALARM_1 = "alarm_1"
ALARM_2 = "alarm_2"
ALARM_3 = "alarm_3"
CAPBANK_RUNNING_TIME = "capbank_running_time"
INTERNAL_FAN_1_RUNNING_TIME = "internal_fan_1_running_time"
INV_MODULE_A_TEMP = "inv_module_a_temp"
INV_MODULE_B_TEMP = "inv_module_b_temp"
INV_MODULE_C_TEMP = "inv_module_c_temp"
ANTI_REVERSE_MODULE_1_TEMP = "anti_reverse_module_1_temp"
ANTI_REVERSE_MODULE_2_TEMP = "anti_reverse_module_2_temp"
OUTPUT_BOARD_RELAY_AMBIENT_TEMP_MAX = "output_board_relay_ambient_temp_max"
DC_TERMINAL_1_2_MAX_TEMP = "dc_terminal_1_2_max_temp"
AC_TERMINAL_1_2_3_MAX_TEMP = "ac_terminal_1_2_3_max_temp"
PHASE_A_DC_COMPONENT_DCI = "phase_a_dc_component_dci"
PHASE_B_DC_COMPONENT_DCI = "phase_b_dc_component_dci"
PHASE_C_DC_COMPONENT_DCI = "phase_c_dc_component_dci"
LEAKAGE_CURRENT_RCD = "leakage_current_rcd"
POSITIVE_BUS_VOLTAGE = "positive_bus_voltage"
NEGATIVE_BUS_VOLTAGE = "negative_bus_voltage"
BUS_NEGATIVE_VOLTAGE_TO_GROUND = "bus_negative_voltage_to_ground"
PV_01_VOLTAGE = "pv_01_voltage"
PV_01_CURRENT = "pv_01_current"
PV_02_VOLTAGE = "pv_02_voltage"
PV_02_CURRENT = "pv_02_current"
PV_03_VOLTAGE = "pv_03_voltage"
PV_03_CURRENT = "pv_03_current"
PV_04_VOLTAGE = "pv_04_voltage"
PV_04_CURRENT = "pv_04_current"
PV_05_VOLTAGE = "pv_05_voltage"
PV_05_CURRENT = "pv_05_current"
PV_06_VOLTAGE = "pv_06_voltage"
PV_06_CURRENT = "pv_06_current"
PV_07_VOLTAGE = "pv_07_voltage"
PV_07_CURRENT = "pv_07_current"
PV_08_VOLTAGE = "pv_08_voltage"
PV_08_CURRENT = "pv_08_current"
PV_09_VOLTAGE = "pv_09_voltage"
PV_09_CURRENT = "pv_09_current"
PV_10_VOLTAGE = "pv_10_voltage"
PV_10_CURRENT = "pv_10_current"
PV_11_VOLTAGE = "pv_11_voltage"
PV_11_CURRENT = "pv_11_current"
PV_12_VOLTAGE = "pv_12_voltage"
PV_12_CURRENT = "pv_12_current"
PV_13_VOLTAGE = "pv_13_voltage"
PV_13_CURRENT = "pv_13_current"
PV_14_VOLTAGE = "pv_14_voltage"
PV_14_CURRENT = "pv_14_current"
PV_15_VOLTAGE = "pv_15_voltage"
PV_15_CURRENT = "pv_15_current"
PV_16_VOLTAGE = "pv_16_voltage"
PV_16_CURRENT = "pv_16_current"
PV_17_VOLTAGE = "pv_17_voltage"
PV_17_CURRENT = "pv_17_current"
PV_18_VOLTAGE = "pv_18_voltage"
PV_18_CURRENT = "pv_18_current"
PV_19_VOLTAGE = "pv_19_voltage"
PV_19_CURRENT = "pv_19_current"
PV_20_VOLTAGE = "pv_20_voltage"
PV_20_CURRENT = "pv_20_current"
PV_21_VOLTAGE = "pv_21_voltage"
PV_21_CURRENT = "pv_21_current"
PV_22_VOLTAGE = "pv_22_voltage"
PV_22_CURRENT = "pv_22_current"
PV_23_VOLTAGE = "pv_23_voltage"
PV_23_CURRENT = "pv_23_current"
PV_24_VOLTAGE = "pv_24_voltage"
PV_24_CURRENT = "pv_24_current"
INPUT_POWER = "input_power"
GRID_VOLTAGE = "grid_voltage"
LINE_VOLTAGE_A_B = "line_voltage_A_B"
LINE_VOLTAGE_B_C = "line_voltage_B_C"
LINE_VOLTAGE_C_A = "line_voltage_C_A"
PHASE_A_VOLTAGE = "phase_A_voltage"
PHASE_B_VOLTAGE = "phase_B_voltage"
PHASE_C_VOLTAGE = "phase_C_voltage"
GRID_CURRENT = "grid_current"
PHASE_A_CURRENT = "phase_A_current"
PHASE_B_CURRENT = "phase_B_current"
PHASE_C_CURRENT = "phase_C_current"
DAY_ACTIVE_POWER_PEAK = "day_active_power_peak"
ACTIVE_POWER = "active_power"
REACTIVE_POWER = "reactive_power"
POWER_FACTOR = "power_factor"
GRID_FREQUENCY = "grid_frequency"
EFFICIENCY = "efficiency"
INTERNAL_TEMPERATURE = "internal_temperature"
INSULATION_RESISTANCE = "insulation_resistance"
DEVICE_STATUS = "device_status"
FAULT_CODE = "fault_code"
STARTUP_TIME = "startup_time"
SHUTDOWN_TIME = "shutdown_time"
ACTIVE_POWER_FAST = "active_power_fast"
ACCUMULATED_YIELD_ENERGY = "accumulated_yield_energy"
TOTAL_DC_INPUT_POWER = "total_dc_input_power"
CURRENT_ELECTRICITY_GENERATION_STATISTICS_TIME = "current_electricity_generation_statistics_time"
HOURLY_YIELD_ENERGY = "hourly_yield_energy"
DAILY_YIELD_ENERGY = "daily_yield_energy"
MONTHLY_YIELD_ENERGY = "monthly_yield_energy"
YEARLY_YIELD_ENERGY = "yearly_yield_energy"
LATEST_ACTIVE_ALARM_SN = "latest_active_alarm_sn"
LATEST_HISTORICAL_ALARM_SN = "latest_historical_alarm_sn"
TOTAL_BUS_VOLTAGE = "total_bus_voltage"
MAX_PV_VOLTAGE = "max_pv_voltage"
MIN_PV_VOLTAGE = "min_pv_voltage"
AVERAGE_PV_NEGATIVE_VOLTAGE_TO_GROUND = "average_pv_negative_voltage_to_ground"
MIN_PV_NEGATIVE_VOLTAGE_TO_GROUND = "min_pv_negative_voltage_to_ground"
MAX_PV_NEGATIVE_VOLTAGE_TO_GROUND = "max_pv_negative_voltage_to_ground"
INVERTER_TO_PE_VOLTAGE_TOLERANCE = "inverter_to_pe_voltage_tolerance"
ISO_FEATURE_INFORMATION = "iso_feature_information"
BUILTIN_PID_RUNNING_STATUS = "builtin_pid_running_status"
PV_NEGATIVE_VOLTAGE_TO_GROUND = "pv_negative_voltage_to_ground"
CUMULATIVE_DC_ENERGY_YIELD_MPPT1 = "cumulative_dc_energy_yield_mppt1"
CUMULATIVE_DC_ENERGY_YIELD_MPPT2 = "cumulative_dc_energy_yield_mppt2"
CUMULATIVE_DC_ENERGY_YIELD_MPPT3 = "cumulative_dc_energy_yield_mppt3"
CUMULATIVE_DC_ENERGY_YIELD_MPPT4 = "cumulative_dc_energy_yield_mppt4"
CUMULATIVE_DC_ENERGY_YIELD_MPPT5 = "cumulative_dc_energy_yield_mppt5"
CUMULATIVE_DC_ENERGY_YIELD_MPPT6 = "cumulative_dc_energy_yield_mppt6"
CUMULATIVE_DC_ENERGY_YIELD_MPPT7 = "cumulative_dc_energy_yield_mppt7"
CUMULATIVE_DC_ENERGY_YIELD_MPPT8 = "cumulative_dc_energy_yield_mppt8"
CUMULATIVE_DC_ENERGY_YIELD_MPPT9 = "cumulative_dc_energy_yield_mppt9"
CUMULATIVE_DC_ENERGY_YIELD_MPPT10 = "cumulative_dc_energy_yield_mppt10"
STORAGE_UNIT_1_RUNNING_STATUS = "storage_unit_1_running_status"
STORAGE_UNIT_1_CHARGE_DISCHARGE_POWER = "storage_unit_1_charge_discharge_power"
STORAGE_UNIT_1_BUS_VOLTAGE = "storage_unit_1_bus_voltage"
STORAGE_UNIT_1_STATE_OF_CAPACITY = "storage_unit_1_state_of_capacity"
STORAGE_UNIT_1_WORKING_MODE_B = "storage_unit_1_working_mode_b"
STORAGE_UNIT_1_RATED_CHARGE_POWER = "storage_unit_1_rated_charge_power"
STORAGE_UNIT_1_RATED_DISCHARGE_POWER = "storage_unit_1_rated_discharge_power"
STORAGE_UNIT_1_FAULT_ID = "storage_unit_1_fault_id"
STORAGE_UNIT_1_CURRENT_DAY_CHARGE_CAPACITY = "storage_unit_1_current_day_charge_capacity"
STORAGE_UNIT_1_CURRENT_DAY_DISCHARGE_CAPACITY = "storage_unit_1_current_day_discharge_capacity"
STORAGE_UNIT_1_BUS_CURRENT = "storage_unit_1_bus_current"
STORAGE_UNIT_1_BATTERY_TEMPERATURE = "storage_unit_1_battery_temperature"
STORAGE_UNIT_1_REMAINING_CHARGE_DIS_CHARGE_TIME = "storage_unit_1_remaining_charge_dis_charge_time"
STORAGE_UNIT_1_DCDC_VERSION = "storage_unit_1_dcdc_version"
STORAGE_UNIT_1_BMS_VERSION = "storage_unit_1_bms_version"
STORAGE_MAXIMUM_CHARGE_POWER = "storage_maximum_charge_power"
STORAGE_MAXIMUM_DISCHARGE_POWER = "storage_maximum_discharge_power"
STORAGE_UNIT_1_SERIAL_NUMBER = "storage_unit_1_serial_number"
STORAGE_UNIT_1_TOTAL_CHARGE = "storage_unit_1_total_charge"
STORAGE_UNIT_1_TOTAL_DISCHARGE = "storage_unit_1_total_discharge"
METER_STATUS = "meter_status"
GRID_A_VOLTAGE = "grid_A_voltage"
GRID_B_VOLTAGE = "grid_B_voltage"
GRID_C_VOLTAGE = "grid_C_voltage"
ACTIVE_GRID_A_CURRENT = "active_grid_A_current"
ACTIVE_GRID_B_CURRENT = "active_grid_B_current"
ACTIVE_GRID_C_CURRENT = "active_grid_C_current"
POWER_METER_ACTIVE_POWER = "power_meter_active_power"
POWER_METER_REACTIVE_POWER = "power_meter_reactive_power"
ACTIVE_GRID_POWER_FACTOR = "active_grid_power_factor"
ACTIVE_GRID_FREQUENCY = "active_grid_frequency"
GRID_EXPORTED_ENERGY = "grid_exported_energy"
GRID_ACCUMULATED_ENERGY = "grid_accumulated_energy"
GRID_ACCUMULATED_REACTIVE_POWER = "grid_accumulated_reactive_power"
METER_TYPE = "meter_type"
ACTIVE_GRID_A_B_VOLTAGE = "active_grid_A_B_voltage"
ACTIVE_GRID_B_C_VOLTAGE = "active_grid_B_C_voltage"
ACTIVE_GRID_C_A_VOLTAGE = "active_grid_C_A_voltage"
ACTIVE_GRID_A_POWER = "active_grid_A_power"
ACTIVE_GRID_B_POWER = "active_grid_B_power"
ACTIVE_GRID_C_POWER = "active_grid_C_power"
NB_OPTIMIZERS = "nb_optimizers"
METER_TYPE_CHECK = "meter_type_check"
NB_ONLINE_OPTIMIZERS = "nb_online_optimizers"
STORAGE_UNIT_2_SERIAL_NUMBER = "storage_unit_2_serial_number"
STORAGE_UNIT_2_STATE_OF_CAPACITY = "storage_unit_2_state_of_capacity"
STORAGE_UNIT_2_RUNNING_STATUS = "storage_unit_2_running_status"
STORAGE_UNIT_2_CHARGE_DISCHARGE_POWER = "storage_unit_2_charge_discharge_power"
STORAGE_UNIT_2_CURRENT_DAY_CHARGE_CAPACITY = "storage_unit_2_current_day_charge_capacity"
STORAGE_UNIT_2_CURRENT_DAY_DISCHARGE_CAPACITY = "storage_unit_2_current_day_discharge_capacity"
STORAGE_UNIT_2_BUS_VOLTAGE = "storage_unit_2_bus_voltage"
STORAGE_UNIT_2_BUS_CURRENT = "storage_unit_2_bus_current"
STORAGE_UNIT_2_BATTERY_TEMPERATURE = "storage_unit_2_battery_temperature"
STORAGE_UNIT_2_TOTAL_CHARGE = "storage_unit_2_total_charge"
STORAGE_UNIT_2_TOTAL_DISCHARGE = "storage_unit_2_total_discharge"
STORAGE_RATED_CAPACITY = "storage_rated_capacity"
STORAGE_STATE_OF_CAPACITY = "storage_state_of_capacity"
STORAGE_RUNNING_STATUS = "storage_running_status"
STORAGE_BUS_VOLTAGE = "storage_bus_voltage"
STORAGE_BUS_CURRENT = "storage_bus_current"
STORAGE_CHARGE_DISCHARGE_POWER = "storage_charge_discharge_power"
STORAGE_TOTAL_CHARGE = "storage_total_charge"
STORAGE_TOTAL_DISCHARGE = "storage_total_discharge"
STORAGE_CURRENT_DAY_CHARGE_CAPACITY = "storage_current_day_charge_capacity"
STORAGE_CURRENT_DAY_DISCHARGE_CAPACITY = "storage_current_day_discharge_capacity"
STORAGE_UNIT_2_SOFTWARE_VERSION = "storage_unit_2_software_version"
STORAGE_UNIT_1_SOFTWARE_VERSION = "storage_unit_1_software_version"
STORAGE_UNIT_1_BATTERY_PACK_1_SOH_CALIBRATION_STATUS = "storage_unit_1_battery_pack_1_soh_calibration_status"
STORAGE_UNIT_1_BATTERY_PACK_2_SOH_CALIBRATION_STATUS = "storage_unit_1_battery_pack_2_soh_calibration_status"
STORAGE_UNIT_1_BATTERY_PACK_3_SOH_CALIBRATION_STATUS = "storage_unit_1_battery_pack_3_soh_calibration_status"
STORAGE_UNIT_2_BATTERY_PACK_1_SOH_CALIBRATION_STATUS = "storage_unit_2_battery_pack_1_soh_calibration_status"
STORAGE_UNIT_2_BATTERY_PACK_2_SOH_CALIBRATION_STATUS = "storage_unit_2_battery_pack_2_soh_calibration_status"
STORAGE_UNIT_2_BATTERY_PACK_3_SOH_CALIBRATION_STATUS = "storage_unit_2_battery_pack_3_soh_calibration_status"
STORAGE_UNIT_SOH_CALIBRATION_STATUS = "storage_unit_soh_calibration_status"
STORAGE_UNIT_SOH_CALIBRATION_RELEASE_LOWER_LIMIT_OF_SOC = "storage_unit_soh_calibration_release_lower_limit_of_soc"
STORAGE_UNIT_1_BATTERY_PACK_1_SERIAL_NUMBER = "storage_unit_1_battery_pack_1_serial_number"
STORAGE_UNIT_1_BATTERY_PACK_1_FIRMWARE_VERSION = "storage_unit_1_battery_pack_1_firmware_version"
STORAGE_UNIT_1_BATTERY_PACK_1_WORKING_STATUS = "storage_unit_1_battery_pack_1_working_status"
STORAGE_UNIT_1_BATTERY_PACK_1_STATE_OF_CAPACITY = "storage_unit_1_battery_pack_1_state_of_capacity"
STORAGE_UNIT_1_BATTERY_PACK_1_CHARGE_DISCHARGE_POWER = "storage_unit_1_battery_pack_1_charge_discharge_power"
STORAGE_UNIT_1_BATTERY_PACK_1_VOLTAGE = "storage_unit_1_battery_pack_1_voltage"
STORAGE_UNIT_1_BATTERY_PACK_1_CURRENT = "storage_unit_1_battery_pack_1_current"
STORAGE_UNIT_1_BATTERY_PACK_1_TOTAL_CHARGE = "storage_unit_1_battery_pack_1_total_charge"
STORAGE_UNIT_1_BATTERY_PACK_1_TOTAL_DISCHARGE = "storage_unit_1_battery_pack_1_total_discharge"
STORAGE_UNIT_1_BATTERY_PACK_2_SERIAL_NUMBER = "storage_unit_1_battery_pack_2_serial_number"
STORAGE_UNIT_1_BATTERY_PACK_2_FIRMWARE_VERSION = "storage_unit_1_battery_pack_2_firmware_version"
STORAGE_UNIT_1_BATTERY_PACK_2_WORKING_STATUS = "storage_unit_1_battery_pack_2_working_status"
STORAGE_UNIT_1_BATTERY_PACK_2_STATE_OF_CAPACITY = "storage_unit_1_battery_pack_2_state_of_capacity"
STORAGE_UNIT_1_BATTERY_PACK_2_CHARGE_DISCHARGE_POWER = "storage_unit_1_battery_pack_2_charge_discharge_power"
STORAGE_UNIT_1_BATTERY_PACK_2_VOLTAGE = "storage_unit_1_battery_pack_2_voltage"
STORAGE_UNIT_1_BATTERY_PACK_2_CURRENT = "storage_unit_1_battery_pack_2_current"
STORAGE_UNIT_1_BATTERY_PACK_2_TOTAL_CHARGE = "storage_unit_1_battery_pack_2_total_charge"
STORAGE_UNIT_1_BATTERY_PACK_2_TOTAL_DISCHARGE = "storage_unit_1_battery_pack_2_total_discharge"
STORAGE_UNIT_1_BATTERY_PACK_3_SERIAL_NUMBER = "storage_unit_1_battery_pack_3_serial_number"
STORAGE_UNIT_1_BATTERY_PACK_3_FIRMWARE_VERSION = "storage_unit_1_battery_pack_3_firmware_version"
STORAGE_UNIT_1_BATTERY_PACK_3_WORKING_STATUS = "storage_unit_1_battery_pack_3_working_status"
STORAGE_UNIT_1_BATTERY_PACK_3_STATE_OF_CAPACITY = "storage_unit_1_battery_pack_3_state_of_capacity"
STORAGE_UNIT_1_BATTERY_PACK_3_CHARGE_DISCHARGE_POWER = "storage_unit_1_battery_pack_3_charge_discharge_power"
STORAGE_UNIT_1_BATTERY_PACK_3_VOLTAGE = "storage_unit_1_battery_pack_3_voltage"
STORAGE_UNIT_1_BATTERY_PACK_3_CURRENT = "storage_unit_1_battery_pack_3_current"
STORAGE_UNIT_1_BATTERY_PACK_3_TOTAL_CHARGE = "storage_unit_1_battery_pack_3_total_charge"
STORAGE_UNIT_1_BATTERY_PACK_3_TOTAL_DISCHARGE = "storage_unit_1_battery_pack_3_total_discharge"
STORAGE_UNIT_2_BATTERY_PACK_1_SERIAL_NUMBER = "storage_unit_2_battery_pack_1_serial_number"
STORAGE_UNIT_2_BATTERY_PACK_1_FIRMWARE_VERSION = "storage_unit_2_battery_pack_1_firmware_version"
STORAGE_UNIT_2_BATTERY_PACK_1_WORKING_STATUS = "storage_unit_2_battery_pack_1_working_status"
STORAGE_UNIT_2_BATTERY_PACK_1_STATE_OF_CAPACITY = "storage_unit_2_battery_pack_1_state_of_capacity"
STORAGE_UNIT_2_BATTERY_PACK_1_CHARGE_DISCHARGE_POWER = "storage_unit_2_battery_pack_1_charge_discharge_power"
STORAGE_UNIT_2_BATTERY_PACK_1_VOLTAGE = "storage_unit_2_battery_pack_1_voltage"
STORAGE_UNIT_2_BATTERY_PACK_1_CURRENT = "storage_unit_2_battery_pack_1_current"
STORAGE_UNIT_2_BATTERY_PACK_1_TOTAL_CHARGE = "storage_unit_2_battery_pack_1_total_charge"
STORAGE_UNIT_2_BATTERY_PACK_1_TOTAL_DISCHARGE = "storage_unit_2_battery_pack_1_total_discharge"
STORAGE_UNIT_2_BATTERY_PACK_2_SERIAL_NUMBER = "storage_unit_2_battery_pack_2_serial_number"
STORAGE_UNIT_2_BATTERY_PACK_2_FIRMWARE_VERSION = "storage_unit_2_battery_pack_2_firmware_version"
STORAGE_UNIT_2_BATTERY_PACK_2_WORKING_STATUS = "storage_unit_2_battery_pack_2_working_status"
STORAGE_UNIT_2_BATTERY_PACK_2_STATE_OF_CAPACITY = "storage_unit_2_battery_pack_2_state_of_capacity"
STORAGE_UNIT_2_BATTERY_PACK_2_CHARGE_DISCHARGE_POWER = "storage_unit_2_battery_pack_2_charge_discharge_power"
STORAGE_UNIT_2_BATTERY_PACK_2_VOLTAGE = "storage_unit_2_battery_pack_2_voltage"
STORAGE_UNIT_2_BATTERY_PACK_2_CURRENT = "storage_unit_2_battery_pack_2_current"
STORAGE_UNIT_2_BATTERY_PACK_2_TOTAL_CHARGE = "storage_unit_2_battery_pack_2_total_charge"
STORAGE_UNIT_2_BATTERY_PACK_2_TOTAL_DISCHARGE = "storage_unit_2_battery_pack_2_total_discharge"
STORAGE_UNIT_2_BATTERY_PACK_3_SERIAL_NUMBER = "storage_unit_2_battery_pack_3_serial_number"
STORAGE_UNIT_2_BATTERY_PACK_3_FIRMWARE_VERSION = "storage_unit_2_battery_pack_3_firmware_version"
STORAGE_UNIT_2_BATTERY_PACK_3_WORKING_STATUS = "storage_unit_2_battery_pack_3_working_status"
STORAGE_UNIT_2_BATTERY_PACK_3_STATE_OF_CAPACITY = "storage_unit_2_battery_pack_3_state_of_capacity"
STORAGE_UNIT_2_BATTERY_PACK_3_CHARGE_DISCHARGE_POWER = "storage_unit_2_battery_pack_3_charge_discharge_power"
STORAGE_UNIT_2_BATTERY_PACK_3_VOLTAGE = "storage_unit_2_battery_pack_3_voltage"
STORAGE_UNIT_2_BATTERY_PACK_3_CURRENT = "storage_unit_2_battery_pack_3_current"
STORAGE_UNIT_2_BATTERY_PACK_3_TOTAL_CHARGE = "storage_unit_2_battery_pack_3_total_charge"
STORAGE_UNIT_2_BATTERY_PACK_3_TOTAL_DISCHARGE = "storage_unit_2_battery_pack_3_total_discharge"
STORAGE_UNIT_1_BATTERY_PACK_1_MAXIMUM_TEMPERATURE = "storage_unit_1_battery_pack_1_maximum_temperature"
STORAGE_UNIT_1_BATTERY_PACK_1_MINIMUM_TEMPERATURE = "storage_unit_1_battery_pack_1_minimum_temperature"
STORAGE_UNIT_1_BATTERY_PACK_2_MAXIMUM_TEMPERATURE = "storage_unit_1_battery_pack_2_maximum_temperature"
STORAGE_UNIT_1_BATTERY_PACK_2_MINIMUM_TEMPERATURE = "storage_unit_1_battery_pack_2_minimum_temperature"
STORAGE_UNIT_1_BATTERY_PACK_3_MAXIMUM_TEMPERATURE = "storage_unit_1_battery_pack_3_maximum_temperature"
STORAGE_UNIT_1_BATTERY_PACK_3_MINIMUM_TEMPERATURE = "storage_unit_1_battery_pack_3_minimum_temperature"
STORAGE_UNIT_2_BATTERY_PACK_1_MAXIMUM_TEMPERATURE = "storage_unit_2_battery_pack_1_maximum_temperature"
STORAGE_UNIT_2_BATTERY_PACK_1_MINIMUM_TEMPERATURE = "storage_unit_2_battery_pack_1_minimum_temperature"
STORAGE_UNIT_2_BATTERY_PACK_2_MAXIMUM_TEMPERATURE = "storage_unit_2_battery_pack_2_maximum_temperature"
STORAGE_UNIT_2_BATTERY_PACK_2_MINIMUM_TEMPERATURE = "storage_unit_2_battery_pack_2_minimum_temperature"
STORAGE_UNIT_2_BATTERY_PACK_3_MAXIMUM_TEMPERATURE = "storage_unit_2_battery_pack_3_maximum_temperature"
STORAGE_UNIT_2_BATTERY_PACK_3_MINIMUM_TEMPERATURE = "storage_unit_2_battery_pack_3_minimum_temperature"
SYSTEM_TIME = "system_time"
SYSTEM_TIME_RAW = "system_time_raw"
Q_U_CHARACTERISTIC_CURVE_MODEL = "q_u_characteristic_curve_model"
Q_U_SCHEDULING_TRIGGER_POWER_PERCENTAGE = "q_u_scheduling_trigger_power_percentage"
POWER_FACTOR_2 = "power_factor_2"
REACTIVE_POWER_COMPENSATION = "reactive_power_compensation"
REACTIVE_POWER_ADJUSTMENT_TIME = "reactive_power_adjustment_time"
ACTIVE_POWER_PERCENTAGE_DERATING = "active_power_percentage_derating"
ACTIVE_POWER_FIXED_VALUE_DERATING = "active_power_fixed_value_derating"
REACTIVE_POWER_COMPENSATION_AT_NIGHT = "reactive_power_compensation_at_night"
FIXED_REACTIVE_POWER_AT_NIGHT = "fixed_reactive_power_at_night"
CHARACTERISTIC_CURVE_REACTIVE_POWER_ADJUSTMENT_TIME = "characteristic_curve_reactive_power_adjustment_time"
PERCENT_APPARENT_POWER = "percent_apparent_power"
Q_U_SCHEDULING_EXIT_POWER_PERCENTAGE = "q_u_scheduling_exit_power_percentage"
STARTUP = "startup"
SHUTDOWN = "shutdown"
GRID_CODE = "grid_code"
MPPT_MULTIMODAL_SCANNING = "mppt_multimodal_scanning"
MPPT_SCANNING_INTERVAL = "mppt_scanning_interval"
MPPT_PREDICTED_POWER = "mppt_predicted_power"
TIME_ZONE = "time_zone"

WLAN_WAKEUP = "wlan_wakeup"

STORAGE_UNIT_1_PRODUCT_MODEL = "storage_unit_1_product_model"
STORAGE_WORKING_MODE_A = "storage_working_mode_a"
STORAGE_TIME_OF_USE_PRICE = "storage_time_of_use_price"
STORAGE_LG_RESU_TIME_OF_USE_PRICE_PERIODS = "storage_lg_resu_time_of_use_price_periods"
STORAGE_HUAWEI_LUNA2000_TIME_OF_USE_PRICE_PERIODS = "storage_huawei_luna2000_time_of_use_price_periods"
STORAGE_LCOE = "storage_lcoe"
STORAGE_MAXIMUM_CHARGING_POWER = "storage_maximum_charging_power"
STORAGE_MAXIMUM_DISCHARGING_POWER = "storage_maximum_discharging_power"
STORAGE_POWER_LIMIT_GRID_TIED_POINT = "storage_power_limit_grid_tied_point"
STORAGE_CHARGING_CUTOFF_CAPACITY = "storage_charging_cutoff_capacity"
STORAGE_DISCHARGING_CUTOFF_CAPACITY = "storage_discharging_cutoff_capacity"
STORAGE_FORCED_CHARGING_AND_DISCHARGING_PERIOD = "storage_forced_charging_and_discharging_period"
STORAGE_FORCED_CHARGING_AND_DISCHARGING_POWER = "storage_forced_charging_and_discharging_power"
STORAGE_WORKING_MODE_SETTINGS = "storage_working_mode_settings"
STORAGE_CHARGE_FROM_GRID_FUNCTION = "storage_charge_from_grid_function"
STORAGE_GRID_CHARGE_CUTOFF_STATE_OF_CHARGE = "storage_grid_charge_cutoff_state_of_charge"
STORAGE_UNIT_2_PRODUCT_MODEL = "storage_unit_2_product_model"
STORAGE_FORCIBLE_CHARGE_DISCHARGE_WRITE = "forcible_charge_discharge_write"
STORAGE_FORCIBLE_CHARGE_DISCHARGE_SOC = "storage_forcible_charge_discharge_soc"
STORAGE_BACKUP_POWER_STATE_OF_CHARGE = "storage_backup_power_state_of_charge"
STORAGE_UNIT_1_NO = "storage_unit_1_no"
STORAGE_UNIT_2_NO = "storage_unit_2_no"
STORAGE_FIXED_CHARGING_AND_DISCHARGING_PERIODS = "storage_fixed_charging_and_discharging_periods"
STORAGE_POWER_OF_CHARGE_FROM_GRID = "storage_power_of_charge_from_grid"
STORAGE_MAXIMUM_POWER_OF_CHARGE_FROM_GRID = "storage_maximum_power_of_charge_from_grid"
STORAGE_FORCIBLE_CHARGE_DISCHARGE_SETTING_MODE = "storage_forcible_charge_discharge_setting_mode"
STORAGE_FORCIBLE_CHARGE_POWER = "storage_forcible_charge_power"
STORAGE_FORCIBLE_DISCHARGE_POWER = "storage_forcible_discharge_power"
STORAGE_LG_RESU_TIME_OF_USE_CHARGING_AND_DISCHARGING_PERIODS = (
    "storage_lg_resu_time_of_use_charging_and_discharging_periods"
)
STORAGE_HUAWEI_LUNA2000_TIME_OF_USE_CHARGING_AND_DISCHARGING_PERIODS = (
    "storage_huawei_luna2000_time_of_use_charging_and_discharging_periods"
)
STORAGE_EXCESS_PV_ENERGY_USE_IN_TOU = "storage_excess_pv_energy_use_in_tou"
ACTIVE_POWER_CONTROL_MODE = "active_power_control_mode"
MAXIMUM_FEED_GRID_POWER_WATT = "maximum_feed_grid_power_watt"
MAXIMUM_FEED_GRID_POWER_PERCENT = "maximum_feed_grid_power_percent"
DONGLE_PLANT_MAXIMUM_CHARGE_FROM_GRID_POWER = "dongle_plant_maximum_charge_from_grid_power"
BACKUP_SWITCH_TO_OFF_GRID = "backup_switch_to_off_grid"
BACKUP_VOLTAGE_INDEPENDENT_OPERATION = "backup_voltage_independent_operation"
STORAGE_UNIT_1_PACK_1_NO = "storage_unit_1_pack_1_no"
STORAGE_UNIT_1_PACK_2_NO = "storage_unit_1_pack_2_no"
STORAGE_UNIT_1_PACK_3_NO = "storage_unit_1_pack_3_no"
STORAGE_UNIT_2_PACK_1_NO = "storage_unit_2_pack_1_no"
STORAGE_UNIT_2_PACK_2_NO = "storage_unit_2_pack_2_no"
STORAGE_UNIT_2_PACK_3_NO = "storage_unit_2_pack_3_no"

REMOTE_CHARGE_DISCHARGE_CONTROL_MODE = "remote_charge_discharge_control_mode"
DEFAULT_MAXIMUM_FEED_IN_POWER = "default_maximum_feed_in_power"
DEFAULT_ACTIVE_POWER_CHANGE_GRADIENT = "default_active_power_change_gradient"

STORAGE_CAPACITY_CONTROL_MODE = "storage_capacity_control_mode"
STORAGE_CAPACITY_CONTROL_SOC_PEAK_SHAVING = "storage_capacity_control_soc_peak_shaving"
STORAGE_CAPACITY_CONTROL_PERIODS = "storage_capacity_control_periods"

SUN2000_EMMA = "emma"
EMMA_SOFTWARE_VERSION = "emma_software_version"
EMMA_MODEL = "emma_model"

INVERTER_TOTAL_ABSORBED_ENERGY = "inverter_total_absorbed_energy"
ENERGY_CHARGED_TODAY = "energy_charged_today"
TOTAL_CHARGED_ENERGY = "total_charged_energy"
ENERGY_DISCHARGED_TODAY = "energy_discharged_today"
TOTAL_DISCHARGED_ENERGY = "total_discharged_energy"
ESS_CHARGEABLE_ENERGY = "ess_chargeable_energy"
ESS_DISCHARGEABLE_ENERGY = "ess_dischargeable_energy"
RATED_ESS_CAPACITY = "rated_ess_capacity"
CONSUMPTION_TODAY = "consumption_today"
TOTAL_ENERGY_CONSUMPTION = "total_energy_consumption"
FEED_IN_TO_GRID_TODAY = "feed_in_to_grid_today"
TOTAL_FEED_IN_TO_GRID = "total_feed_in_to_grid"
SUPPLY_FROM_GRID_TODAY = "supply_from_grid_today"
TOTAL_SUPPLY_FROM_GRID = "total_supply_from_grid"
INVERTER_ENERGY_YIELD_TODAY = "inverter_energy_yield_today"
INVERTER_TOTAL_ENERGY_YIELD = "inverter_total_energy_yield"
PV_YIELD_TODAY = "pv_yield_today"
TOTAL_PV_ENERGY_YIELD = "total_pv_energy_yield"
PV_OUTPUT_POWER = "pv_output_power"
LOAD_POWER = "load_power"
FEED_IN_POWER = "feed_in_power"
BATTERY_CHARGE_DISCHARGE_POWER = "battery_charge_discharge_power"
INVERTER_RATED_POWER = "inverter_rated_power"
INVERTER_ACTIVE_POWER = "inverter_active_power"
STATE_OF_CAPACITY = "state_of_capacity"
ESS_CHARGEABLE_CAPACITY = "ess_chargeable_capacity"
ESS_DISCHARGEABLE_CAPACITY = "ess_dischargeable_capacity"
BACKUP_POWER_STATE_OF_CHARGE = "backup_power_state_of_charge"
YIELD_THIS_MONTH = "yield_this_month"
MONTHLY_ENERGY_CONSUMPTION = "monthly_energy_consumption"
MONTHLY_FEED_IN_TO_GRID = "monthly_feed_in_to_grid"
YIELD_THIS_YEAR = "yield_this_year"
ANNUAL_ENERGY_CONSUMPTION = "annual_energy_consumption"
YEARLY_FEED_IN_TO_GRID = "yearly_feed_in_to_grid"
MONTHLY_SUPPLY_FROM_GRID = "monthly_supply_from_grid"
YEARLY_SUPPLY_FROM_GRID = "yearly_supply_from_grid"
BACKUP_TIME_NOTIFICATION_THRESHOLD = "backup_time_notification_threshold"
ENERGY_CHARGED_THIS_MONTH = "energy_charged_this_month"
ENERGY_DISCHARGED_THIS_MONTH = "energy_discharged_this_month"
EMMA_EXTERNAL_METER_RUNNING_STATUS = "emma_external_meter_running_status"
EMMA_EXTERNAL_METER_PHASE_A_VOLTAGE = "emma_external_meter_phase_a_voltage"
EMMA_EXTERNAL_METER_PHASE_B_VOLTAGE = "emma_external_meter_phase_b_voltage"
EMMA_EXTERNAL_METER_PHASE_C_VOLTAGE = "emma_external_meter_phase_c_voltage"
EMMA_EXTERNAL_METER_LINE_VOLTAGE_A_B = "emma_external_meter_line_voltage_a_b"
EMMA_EXTERNAL_METER_LINE_VOLTAGE_B_C = "emma_external_meter_line_voltage_b_c"
EMMA_EXTERNAL_METER_LINE_VOLTAGE_C_A = "emma_external_meter_line_voltage_c_a"
EMMA_EXTERNAL_METER_PHASE_A_CURRENT = "emma_external_meter_phase_a_current"
EMMA_EXTERNAL_METER_PHASE_B_CURRENT = "emma_external_meter_phase_b_current"
EMMA_EXTERNAL_METER_PHASE_C_CURRENT = "emma_external_meter_phase_c_current"
EMMA_EXTERNAL_METER_ACTIVE_POWER = "emma_external_meter_active_power"
EMMA_EXTERNAL_METER_POWER_FACTOR = "emma_external_meter_power_factor"
EMMA_EXTERNAL_METER_APPARENT_POWER = "emma_external_meter_apparent_power"
EMMA_EXTERNAL_METER_PHASE_A_ACTIVE_POWER = "emma_external_meter_phase_a_active_power"
EMMA_EXTERNAL_METER_PHASE_B_ACTIVE_POWER = "emma_external_meter_phase_b_active_power"
EMMA_EXTERNAL_METER_PHASE_C_ACTIVE_POWER = "emma_external_meter_phase_c_active_power"
EMMA_EXTERNAL_METER_TOTAL_ACTIVE_ENERGY = "emma_external_meter_total_active_energy"
EMMA_EXTERNAL_METER_TOTAL_NEGATIVE_ACTIVE_ENERGY = "emma_external_meter_total_negative_active_energy"
EMMA_EXTERNAL_METER_TOTAL_POSITIVE_ACTIVE_ENERGY = "emma_external_meter_total_positive_active_energy"
NUMBER_OF_INVERTERS_FOUND = "number_of_inverters_found"
NUMBER_OF_CHARGERS_FOUND = "number_of_chargers_found"
PHASE_A_VOLTAGE_BUILT_IN_ENERGY_SENSOR = "phase_a_voltage_built_in_energy_sensor"
PHASE_B_VOLTAGE_BUILT_IN_ENERGY_SENSOR = "phase_b_voltage_built_in_energy_sensor"
PHASE_C_VOLTAGE_BUILT_IN_ENERGY_SENSOR = "phase_c_voltage_built_in_energy_sensor"
LINE_VOLTAGE_A_B_BUILT_IN_ENERGY_SENSOR = "line_voltage_a_b_built_in_energy_sensor"
LINE_VOLTAGE_B_C_BUILT_IN_ENERGY_SENSOR = "line_voltage_b_c_built_in_energy_sensor"
LINE_VOLTAGE_C_A_BUILT_IN_ENERGY_SENSOR = "line_voltage_c_a_built_in_energy_sensor"
PHASE_A_CURRENT_BUILT_IN_ENERGY_SENSOR = "phase_a_current_built_in_energy_sensor"
PHASE_B_CURRENT_BUILT_IN_ENERGY_SENSOR = "phase_b_current_built_in_energy_sensor"
PHASE_C_CURRENT_BUILT_IN_ENERGY_SENSOR = "phase_c_current_built_in_energy_sensor"
ACTIVE_POWER_BUILT_IN_ENERGY_SENSOR = "active_power_built_in_energy_sensor"
POWER_FACTOR_BUILT_IN_ENERGY_SENSOR = "power_factor_built_in_energy_sensor"
APPARENT_POWER_BUILT_IN_ENERGY_SENSOR = "apparent_power_built_in_energy_sensor"
PHASE_A_ACTIVE_POWER_BUILT_IN_ENERGY_SENSOR = "phase_a_active_power_built_in_energy_sensor"
PHASE_B_ACTIVE_POWER_BUILT_IN_ENERGY_SENSOR = "phase_b_active_power_built_in_energy_sensor"
PHASE_C_ACTIVE_POWER_BUILT_IN_ENERGY_SENSOR = "phase_c_active_power_built_in_energy_sensor"
TOTAL_ACTIVE_ENERGY_BUILT_IN_ENERGY_SENSOR = "total_active_energy_built_in_energy_sensor"
TOTAL_NEGATIVE_ACTIVE_ENERGY_BUILT_IN_ENERGY_SENSOR = "total_negative_active_energy_built_in_energy_sensor"
TOTAL_POSITIVE_ACTIVE_ENERGY_BUILT_IN_ENERGY_SENSOR = "total_positive_active_energy_built_in_energy_sensor"
PHASE_A_VOLTAGE_EXTERNAL_ENERGY_SENSOR = "phase_a_voltage_external_energy_sensor"
PHASE_B_VOLTAGE_EXTERNAL_ENERGY_SENSOR = "phase_b_voltage_external_energy_sensor"
PHASE_C_VOLTAGE_EXTERNAL_ENERGY_SENSOR = "phase_c_voltage_external_energy_sensor"
LINE_VOLTAGE_A_B_EXTERNAL_ENERGY_SENSOR = "line_voltage_a_b_external_energy_sensor"
LINE_VOLTAGE_B_C_EXTERNAL_ENERGY_SENSOR = "line_voltage_b_c_external_energy_sensor"
LINE_VOLTAGE_C_A_EXTERNAL_ENERGY_SENSOR = "line_voltage_c_a_external_energy_sensor"
PHASE_A_CURRENT_EXTERNAL_ENERGY_SENSOR = "phase_a_current_external_energy_sensor"
PHASE_B_CURRENT_EXTERNAL_ENERGY_SENSOR = "phase_b_current_external_energy_sensor"
PHASE_C_CURRENT_EXTERNAL_ENERGY_SENSOR = "phase_c_current_external_energy_sensor"
ACTIVE_POWER_EXTERNAL_ENERGY_SENSOR = "active_power_external_energy_sensor"
POWER_FACTOR_EXTERNAL_ENERGY_SENSOR = "power_factor_external_energy_sensor"
APPARENT_POWER_EXTERNAL_ENERGY_SENSOR = "apparent_power_external_energy_sensor"
PHASE_A_ACTIVE_POWER_EXTERNAL_ENERGY_SENSOR = "phase_a_active_power_external_energy_sensor"
PHASE_B_ACTIVE_POWER_EXTERNAL_ENERGY_SENSOR = "phase_b_active_power_external_energy_sensor"
PHASE_C_ACTIVE_POWER_EXTERNAL_ENERGY_SENSOR = "phase_c_active_power_external_energy_sensor"
TOTAL_ACTIVE_ENERGY_EXTERNAL_ENERGY_SENSOR = "total_active_energy_external_energy_sensor"
TOTAL_NEGATIVE_ACTIVE_ENERGY_EXTERNAL_ENERGY_SENSOR = "total_negative_active_energy_external_energy_sensor"
TOTAL_POSITIVE_ACTIVE_ENERGY_EXTERNAL_ENERGY_SENSOR = "total_positive_active_energy_external_energy_sensor"
EMMA_ESS_CONTROL_MODE = "emma_ess_control_mode"
EMMA_TOU_PREFERRED_USE_OF_SURPLUS_PV_POWER = "emma_tou_preferred_use_of_surplus_pv_power"
EMMA_TOU_MAXIMUM_POWER_FOR_CHARGING_BATTERIES_FROM_GRID = "emma_tou_maximum_power_for_charging_batteries_from_grid"
EMMA_TOU_PERIODS = "emma_tou_periods"
EMMA_POWER_CONTROL_MODE_AT_GRID_CONNECTION_POINT = "emma_power_control_mode_at_grid_connection_point"
EMMA_LIMITATION_MODE = "emma_limitation_mode"
EMMA_MAXIMUM_FEED_GRID_POWER_WATT = "emma_maximum_feed_grid_power_watt"
EMMA_MAXIMUM_FEED_GRID_POWER_PERCENT = "emma_maximum_feed_grid_power_percent"
EMMA_3PHASE_IMBALANCE_CONTROL = "emma_3phase_imbalance_control"
EMMA_POWER_SUPPLY_CONFIGURATION = "emma_power_supply_configuration"
EMMA_CONSIDER_MAINS_FAULTY_IF = "emma_consider_mains_faulty_if"
EMMA_SYSTEM_TIME = "emma_system_time"
LOCAL_TIME_YEAR = "local_time_year"
CHARGER_SOFTWARE_VERSION = "charger_software_version"
CHARGER_ESN = "charger_serial_number"
CHARGER_MODEL = "charger_model"
CHARGER_RATED_POWER = "charger_rated_power"
CHARGER_PHASE_A_VOLTAGE_SENSOR = "charger_phase_a_voltage_sensor"
CHARGER_PHASE_B_VOLTAGE_SENSOR = "charger_phase_b_voltage_sensor"
CHARGER_PHASE_C_VOLTAGE_SENSOR = "charger_phase_c_voltage_sensor"
CHARGER_TOTAL_ENERGY_CHARGED_SENSOR = "chager_total_energy_charged_sensor"
CHARGER_TEMPERATURE_SENSOR = "charger_temperature_sensor"

DAYLIGHT_SAVING_TIME = "daylight_saving_time"
SDONGLE_TOTAL_INPUT_POWER = "sdongle_total_input_power"
SDONGLE_LOAD_POWER = "sdongle_load_power"
SDONGLE_GRID_POWER = "sdongle_grid_power"
SDONGLE_TOTAL_BATTERY_POWER = "sdongle_total_battery_power"
SDONGLE_TOTAL_ACTIVE_POWER = "sdongle_total_active_power"
